import os
import joblib
import importlib
import pandas as pd

from importlib.metadata import version
try:
    __version__ = version('akerbp.models')
except:
   __version__ = 'unknown' 

class Model():
    """    
    The main class to define a model

    Attributes
    ----------
        settings: dict
            the required keys for the dictionary:
                arbitrary model name: dict containing following keys (dict):
                    'model_parameters': dict of parameters required by xgboost.XGBRFClassifier (dict),
                    'use_class_weights': whether to use class weights in the model (boolean)
        model_path: str
            - If model_path is a path/to/dir/ then the serialized model is assumed to be (or will be saved to) 
            path/to/dir/model.joblib, and the labedl encoder at /path/to/dir/le.joblib
            - If model_path is a full/path/to/my_model.joblib,
            then the label encoder is saved (or assumed to be) at /full/path/to/my_model_le.joblib

    Methods
    -------
        predict(X:dict/pandas.DataFrame)

        evaluate(X_test:pandas.DataFrame, y_test:pandas.DataFrame, metrics:list, label_map:dict, **kwargs)

        save(model_path:str)
    """
    def __init__(self, settings, model_path, load=True):
        self.settings = settings
        for key, val in settings.items():
            setattr(self, key, val)

        self.model_path = model_path
        self._handle_model_path()

        if load:
            try:
                self.model = joblib.load(self.model_file_path)
                print('Model successfully loaded from ', self.model_file_path)
                self.le = joblib.load(self.le_file_path)
                print('LabelEncoder successfully loaded from ', self.le_file_path)
            except Exception as e:
                print('Cannot load the model - ', e)

    def _handle_model_path(self):
        if os.path.isfile(self.model_path):
            self.model_file_path = self.model_path
        else:
            if not os.path.isdir(self.model_path):
                os.makedirs(self.model_path)
            self.model_file_path = os.path.join(self.model_path, 'model.joblib')
        self.le_file_path = self.model_file_path.replace('.joblib', '_le.joblib')

    def predict(self, X, **kwarg):
        """
        Parameters
        ----------
            X: dict or pandas DataFrame with expected features

        Returns
        -------
            y: numpy array with predicted classes
        """
        if not hasattr(self, 'model'):
            raise ValueError('Model must be trained or loaded from file before predict can be run')
        self.X_pred = X
        self._validate_x_pred()

        if hasattr(self, '_validate_features'):
            self._validate_features()
        #FIXME! To fix the error that xgboost started throwing on both curve patching and badlog. investigate more?
        feats = self.model.get_booster().feature_names
        predictions = self.model.predict(self.X_pred[feats])
        if hasattr(self, 'le'):
            return self.le.inverse_transform(predictions)
        else:
            return predictions

    def _validate_x_pred(self):
        if isinstance(self.X_pred, dict):
            self.X_pred = pd.DataFrame.from_dict(self.X_pred)
        elif isinstance(self.X_pred, pd.core.frame.DataFrame):
            self.X_pred = self.X_pred
        else:
            raise ValueError('Please pass the data as a dict or a pandas DataFrame')


    def evaluate(self, X_test, y_test, metrics=None, label_map=None, **kwargs):
        """
        Evaluates the metrics for the test data

        Parameters
        ----------
        X_test: pandas DataFrame - raw features
        y_test: pandas DataFrame - correct labels
        metrics(optional): list - evaluation metrics you want calculated from the methods available at sklearn.metrics
            if no list is provided, the metrics list from the settings is used
        label_map(optional): dict - mapping the test set labels to another set, e.g. for metrics with this option
        **kwargs: arbitrary arguments that the expected metrics may need
        """
        if metrics is not None:
            self.metrics = metrics
        self.scores = dict()
        for metric in self.metrics:
            module_name = 'sklearn.metrics'
            cls_name = metric
            try:
                cls = getattr(importlib.import_module(module_name), cls_name)
            except AttributeError as ae:
                print(ae)
            else:
                y_pred = pd.DataFrame()
                y_pred['true'] = pd.to_numeric(y_test)
                y_pred['pred'] = pd.to_numeric(self.predict(X=X_test))
                if label_map is not None:                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        
                    y_pred['pred'] = y_pred['pred'].map(label_map)
                    y_pred['true'] = y_pred['true'].map(label_map)                    
                score = cls(y_pred['true'], y_pred['pred'], **kwargs)
                setattr(self, metric, score)
            self.scores[metric] = getattr(self, metric)
        return self.scores

    def save(self, model_path=None):
        if model_path is not None:
            self.model_path = model_path            

        self._handle_model_path()
        joblib.dump(self.model, self.model_file_path)
        print('Model successfully saved to {}'.format(self.model_file_path))
        joblib.dump(self.le, self.le_file_path)
        print('LabelEncoder successfully saved to {}'.format(self.le_file_path))
        return model_path


import numpy as np
from akerbp.models.rule_based_models import BS_helpers
from akerbp.models.rule_based_models import crossplot_helpers

def flag_dencorr(df_well, y_pred=None, **kwargs):
    """
    Returns anomalous density corrections

    Args:
        df_well (pd.DataFrame): [description]

    Returns:
        [type]: [description]
    """
    print('Method: Density correction...')
    if y_pred is None:
        y_pred = df_well.copy()

    x = 'DEN'
    y = 'DENC'
    df_well = BS_helpers.find_BS_jumps(df_well).copy() 

    dencorr_anomalies = []
    for bsr in df_well['BS_region'].unique():
        df_gr = df_well[df_well['BS_region']==bsr].copy()
        unacceptable_denc = set(df_gr.loc[np.abs(df_gr['DENC']>0.1)].index.tolist())
        if (df_gr[x].dropna().shape[0] != 0) and (df_gr[y].dropna().shape[0] != 0):
            anomalies, scores, idx = crossplot_helpers.find_crossplot_scores(df_gr, x=x, y=y, **kwargs)
            #post process the results
            dencorr_anomalies.extend(list(set(anomalies).intersection(unacceptable_denc)))
            
            #scores
            for a_method, sc in scores.items():
                y_pred.loc[idx, '{}_dencorr'.format(a_method)] = sc
                y_pred['{}_dencorr'.format(a_method)].fillna(0, inplace=True)
                #post process the results
                y_pred.loc[unacceptable_denc, '{}_dencorr'.format(a_method)] = 1
    y_pred.loc[:, ['flag_dencorr_gen', 'flag_dencorr_den']] = 0, 0
    y_pred.loc[dencorr_anomalies, ['flag_dencorr_gen', 'flag_dencorr_den']] = 1
    return y_pred


import pandas as pd
from sklearn.preprocessing import MinMaxScaler

### Correlations
def get_low_pearson_r(df_well, r_window_size, logname, col, thresh):
    """
    Flags samples with low correlation compared to window around it

    Args:
        df_well (pd.DataFrame): data of one well
        r_window_size (int): window size
        logname (str): log being analyzed
        col (str): column being compared to
        thresh (float): threshold for being a bad correlation

    Returns:
        pd.DataFrame: flagged dataframe
    """
    df_interpolated  = df_well[[logname, col]].interpolate()
    rolling_r        = df_interpolated[logname].rolling(
        window=r_window_size,
        center=True
        ).corr(df_interpolated[col])
    df_well['r_{}'.format(col)]      = rolling_r
    df_well['bad_wr_{}'.format(col)] = df_well['r_{}'.format(col)]<thresh
    #Fill the gap between bad samples if two consecutive bad samples are less than "r_window_size" apart
    mask  = df_well['bad_wr_{}'.format(col)].groupby(
        (df_well['bad_wr_{}'.format(col)] != df_well['bad_wr_{}'.format(col)].shift()).cumsum()
        ).transform('count').lt(r_window_size)
    mask &= df_well['bad_wr_{}'.format(col)].eq(False)
    df_well['bad_wr_{}'.format(col)].update(df_well['bad_wr_{}'.format(col)].loc[mask].replace(False, True))
    return df_well

def flag_well(df_well, logname, cols, r_window_size, thresh):
    """
    Flags samples with anomalous correlations

    Args:
        df_well (pd.DataFrame): data of one well
        logname (str): log being analyzed
        cols (list): list of columns to compare logname to
        r_window_size (int): window size
        thresh ([type]): [description]

    Returns:
        pd.DataFrame: flagged dataframe
    """
    df_well_flagged    = pd.DataFrame()
    df_well[[logname]] = MinMaxScaler().fit_transform(df_well[[logname]].values)
    df_well['bad_r']   = False
    for col in cols:
        df_well[[col]]   = MinMaxScaler().fit_transform(df_well[[logname]].values)
        df_well          = get_low_pearson_r(df_well, r_window_size, logname, col, thresh)
        df_well['bad_r'] = df_well['bad_wr_{}'.format(col)] | df_well['bad_r']
    df_well_flagged = df_well_flagged.append(df_well)
    return df_well_flagged

def get_corr_anomalies(df_well, logname, cols, r_window_size, thresh):
    """
    Flags correlations anomalies

    Args:
        df_well (pd.DataFrame): data of one well
        logname (str): log being analyzed
        cols (list): list of columns to compare logname to
        r_window_size (int): window size
        thresh ([type]): [description]

    Returns:
        list: list of anomalous indices
    """
    df_well.fillna(method='ffill', inplace=True)
    df_well.fillna(method='bfill', inplace=True)
    df_flagged = flag_well(df_well, logname, [c for c in cols if c!=logname], r_window_size, thresh)
    df_flagged.fillna(False, inplace=True)
    return df_flagged[df_flagged.bad_r==True].index

def flag_correlate(df_well, y_pred=None, **kwargs):
    """
    Returns anomalous correlation indices for AC, ACS and DEN

    Args:
        df_well ([type]): [description]

    Returns:
        [type]: [description]
    """
    print('Method: correlation...')
    if y_pred is None:
        y_pred = df_well.copy()

    expected_curves = kwargs.get('expected_curves', set(['DEPTH', 'DEN', 'DENC', 'AC', 'ACS', 'BS', 'CALI', 'GR', 'NEU', 'RDEP', 'RMED', 'RMIC', 'GROUP']))
    y_pred.loc[:, ['flag_correlation_gen', 'flag_correlation_ac', 'flag_correlation_acs', 'flag_correlation_den']] =\
        0, 0, 0, 0

    r_window_size      = kwargs.get('r_window_size', 5)
    thresh             = kwargs.get('thresh', 1)
    cols               = kwargs.get('cols', list(expected_curves))
    ac_corr_anomalies  = get_corr_anomalies(df_well, 'AC', cols, r_window_size, thresh)
    acs_corr_anomalies = get_corr_anomalies(df_well, 'ACS', cols, r_window_size, thresh)
    den_corr_anomalies = get_corr_anomalies(df_well, 'DEN', cols, r_window_size, thresh)

    y_pred.loc[ac_corr_anomalies, ['flag_correlation_gen', 'flag_correlation_ac']]   = 1
    y_pred.loc[acs_corr_anomalies, ['flag_correlation_gen', 'flag_correlation_acs']] = 1
    y_pred.loc[den_corr_anomalies, ['flag_correlation_gen', 'flag_correlation_den']] = 1

    return y_pred



import os
import pandas as pd

from akerbp.models.classification_models import ClassificationModel
from akerbp.models.classification_models import XGBoostClassificationModel

class HierarchicalModel(ClassificationModel):
    def __init__(self, highlevel_settings, lowlevel_settings, folder_path, mapping):
        """ 
        A specific subclass of XGBoostClassificationModel that trains models in two levels, 
        the _highlevel_ model that is provided with categories of labels, 
        and the _lowlevel_ models that are a set of models trained on more details labels 
        (the number of lowlevel models equals the number of categories of the highlevel model)

        Attributes
        ----------
            - highlevel_settings: dict - setting for the high level model - see settings for the parent class
            - lowlevel_settings: dict - shared setting for low level models - see settings for the parent class
            * Note that there is only one setting (model parameters etc.) is provided for each level, 
            therefore the "num_class" argument is set based on the number of labels available in training data
            - folder_path: str
            - mapping: dict - how to map categories(high level labels) to detailed classes
                keys: categories (labels used in highlevel model)
                values: detailed classes (labels used in corresponding lowlevel models)

        Methods
        -------

        """
        self.cat2cls_mapping = mapping
        self.folder_path = folder_path
        self.highlevel_folder_path = os.path.join(self.folder_path, 'high_level')
        self.lowlevel_folder_path = os.path.join(self.folder_path, 'low_level')
        self.highlevel_model = XGBoostClassificationModel(highlevel_settings, self.highlevel_folder_path)
        self.lowlevel_models = dict()
        for lowlevel_name, _ in self.cat2cls_mapping.items():
            self.lowlevel_models[lowlevel_name] = XGBoostClassificationModel(
                settings=lowlevel_settings,
                model_path=os.path.join(self.lowlevel_folder_path, f'{lowlevel_name}'))

    def train(self, df, targets):
        print('Training the high level model')
        high_level_df, high_level_y = do_sth(df, targets, mapping1)
        self.highlevel_model.train(high_level_df, high_level_y)
        
        print('Training low level models')
        for model_label, lowlevel_model in self.lowlevel_models.items():
            low_level_df, low_level_y = do_sth(df, targets, mapping2, model_label)
            lowlevel_model.train(low_level_df, low_level_y)

    def predict(self, X):
        print('Predicting high level labels')
        if not isinstance(X, pd.DataFrame):
            raise TypeError("Only pandas DataFrame is supported at the moment.")
        y = X.copy()
        y['high_level'] = self.highlevel_model.predict(X)
        print('Predicting low level labels')
        for model_label, lowlevel_model in self.lowlevel_models.items():
            print(model_label)
            filt = y['high_level']==model_label
            try:
                pred = lowlevel_model.predict(y.loc[filt, X.columns])
                y.loc[filt, 'low_level'] = pred
            except ValueError as ve:
                #Unseen labels problem is likely to arise for categories that are made of classes with too few samples
                print(ve)
        return y

    def save(self, folder_path=None):
        if folder_path is None:
            folder_path = self.folder_path

        self.highlevel_model.save(self.highlevel_folder_path)
        for model_label, lowlevel_model in self.lowlevel_models.items():
            full_model_path = os.path.join(self.lowlevel_folder_path, '{}'.format(model_label))
            lowlevel_model.save(full_model_path)
            print('model {} saved in {}'.format(model_label, full_model_path))



import os
import joblib
import pandas as pd
import importlib

from mlpet.Datasets import lithology
from akerbp.models import Model
from akerbp.models import classification_models
from akerbp.models.rule_based_models import helpers

class BadlogModel(Model):
    def __init__(self, **kwargs):
        """
        Initializes badlog model
        """
        self.kwargs = kwargs
        for key, val in kwargs.items():
            setattr(self, key, val)
        
    def _validate_input(self, X, **kwargs):
        if isinstance(X, dict):
            if 'metadata' not in X.keys():
                raise ValueError('"metadata" key is required')
            if 'data' not in X.keys():
                raise ValueError('"data" key is required')
            X_df = pd.DataFrame.from_dict(X['data'])
            metadata = X['metadata']
            self.X_pred, self.num_cols, self.cat_cols = helpers._apply_metadata(
                X_df, **kwargs.get('_metadata', metadata)
            )
        elif isinstance(X, pd.core.frame.DataFrame):
            X_df = X.copy()
            self.X_pred, self.num_cols, self.cat_cols = helpers._apply_metadata(
                X_df, **kwargs.get('_metadata', {})
            )
        else:
            raise ValueError('Please pass the data as a dict or a pandas DataFrame')

    def _init_cp_models(self, cp_names, folder_path, data_settings, model_settings, mappings):
        """
        Instantiate datasets and models for each of the crossplots

        Args:
            cp_names (list): list with crossplot names
            folder_path (str): path to folder with models
            data_settings (dict): settings of preprocessing for using the model
            model_settings (dict): settings of model XGBoost
            mappings (dict): mappings dictionary of cat variables

        Returns:
            tuple: dictionaries with dataset objects, models and key wells
        """
        datasets = dict()
        models = dict()
        key_wells = dict()
        folder_paths = dict()

        for cp_name in cp_names:
            print('init ds and model - BadlogModel - {}'.format(cp_name))
            folder_paths[cp_name] = os.path.join(folder_path, cp_name)
            datasets[cp_name] = lithology.Lithologydata(
                settings=data_settings, 
                mappings=mappings,
                folder_path=folder_paths[cp_name]
            ) #but don't load any data into the dataset, because we need to get that from running the find_crossplot_scores
            
            models[cp_name] = classification_models.XGBoostClassificationModel(
                model_settings, 
                folder_paths[cp_name]
            )

            key_wells[cp_name] = joblib.load(
                os.path.join(folder_paths[cp_name], 'key_wells.joblib')
            )
        return datasets, models, key_wells

    def _get_cp_params(self, **kwargs):
        """
        Returns necessary variables to generate crossplots predictions

        Returns:
            tuple: path to models folder, settings for data and models, mappings dictionary
        """
        settings = kwargs.get('settings', None)
        mappings = kwargs.get('mappings', None)
        folder_path = kwargs.get('folder_path', None)
        #Assuming all cp_names use the same model- and data-settings
        if all([settings, mappings, folder_path]):
            model_settings = settings['models']
            data_settings = settings['datasets']['lithology']
        else:
            raise ValueError(
                'Please provide settings, mappings and folder path for the crossplot method.'
            )
        return folder_path, data_settings, model_settings, mappings


    def predict(self, X, methods, **kwargs):
        """
        Args:
            X (dict or pd.DataFrame): expected features
            methods (list): list of methods to apply on X

        Returns:
            y (pd.DataFrame): including different badlog flags and scores
        """
        self._validate_input(X, **kwargs)

        #Validate feature
        self.X_pred = helpers._create_features(self.X_pred)

        self.X_pred_features = set(self.X_pred.columns)
        self.expected_curves = set([
            'DEPTH', 'DEN', 'DENC', 'AC', 'ACS', 'BS', 'CALI', 'GR',
            'NEU', 'RDEP', 'RMED', 'RMIC', 'PEF', 'GROUP'
        ])
        helpers._validate_features(self.X_pred_features, self.expected_curves)

        self.methods = methods
        self.y = pd.DataFrame()

        for method in self.methods:
            #Evaluate method
            method_helpers = importlib.import_module(
                'akerbp.models.rule_based_models.{}_helpers'.format(method)
            )
            method_evaluator = getattr(method_helpers, 'flag_{}'.format(method))
            if method == 'crossplot':
                cp_names = kwargs.get('cp_names', ['vp_den', 'vp_vs', 'ai_vpvs'])
                datasets, models, key_wells = self._init_cp_models(
                    cp_names, *self._get_cp_params(**kwargs)
                )
                method_flags = method_evaluator(
                    self.X_pred, cp_names, datasets, models, key_wells
                )
            else:
                method_flags = method_evaluator(self.X_pred)
            self.y[method_flags.columns] = method_flags
        result_columns = [col for col in self.y.columns if col.split('_')[0] in ['flag', 'agg', 'score']]
        return self.y[result_columns]


import numpy as np
import pandas as pd
from sklearn.covariance import EllipticEnvelope
from sklearn.svm import OneClassSVM
from sklearn.ensemble import IsolationForest

from pyod.utils.utility import standardizer #returns z-score


## Crossplot

def get_params(**kwargs):
    """
    Gets parameters for the outlier detection algorithms

    Returns:
        tuple: dictionaries with parameters per method
    """
    SVM_params = {
        'nu': kwargs.get('SVM_nu', 0.01)}
    IsoFor_params = {
        'n_estimators': kwargs.get('IsoFor_n_estimators', 50),
        'contamination': kwargs.get('IsoFor_contamination', 0.01),
        'random_state': kwargs.get('IsoFor_random_state', 0)
    }
    Abod_parameters = {
        'contamination': 0.1,
        'n_neighbors': 10
    }
    EliEnv_parameters = {
        'contamination': 0.05,
        'random_state': 0
    }
    return SVM_params, IsoFor_params, Abod_parameters, EliEnv_parameters

def find_crossplot_scores(df, x, y, algo='comb', **kwargs):
    """
    Find anomalies based on the crossplots of x vs. y

    Args:
        df_well (pd.DataFrame): dataframe with data from one well
        algo (str, optional): which algorithm to use. Defaults to 'comb'.

    Returns:
        list: list with indices of anomalous samples
    """
    algo = algo.lower()
    df  = df[(df[x]!=0) & (df[y]!=0)].copy()
    tmp = df[[x, y]].dropna().copy()

    anomal_idx = []
    method_scores = dict()
    return_scores = dict()
    SVM_params, IsoFor_params, Abod_parameters, EliEnv_parameters = get_params(**kwargs)
    
    #if there are not even 10 samples we return
    if len(tmp)<10:
        return [], {}, tmp.index

    if (algo=='comb') or (algo=='EliEnv'):
        try:
            elienv = EllipticEnvelope(
                contamination = EliEnv_parameters['contamination'],
                random_state = EliEnv_parameters['random_state']
            ).fit(tmp[[x, y]])    
            preds2 = elienv.predict(tmp[[x, y]])
            original_scores = elienv.score_samples(tmp[[x, y]])
        except: #FIXME: elliptic envelope throws covariance error when not so many samples.Quick fix for now.
            preds2 = np.zeros(len(tmp))
            original_scores = np.zeros(len(tmp))
        if algo=='EliEnv':
            preds = preds2
        method_scores['EliEnv'] = np.abs(standardizer(original_scores.reshape(-1,1)))

    if (algo=='comb') or (algo=='svm'):
        svm = OneClassSVM(
            kernel='rbf', 
            gamma='scale', 
            nu=SVM_params['nu']
        ).fit(tmp[[x, y]])
        preds3 = svm.predict(tmp[[x, y]])
        if algo=='svm':
            preds = preds3
        original_scores = svm.score_samples(tmp[[x, y]])
        method_scores['SVM'] = np.abs(standardizer(original_scores.reshape(-1,1)))

    if (algo=='comb') or (algo=='isofor'):
        ifo = IsolationForest(
            n_estimators=IsoFor_params['n_estimators'], 
            contamination=IsoFor_params['contamination'], 
            random_state=IsoFor_params['random_state']
        ).fit(tmp[[x, y]])
        preds4 = ifo.predict(tmp[[x, y]])
        if algo=='isofor':
            preds = preds4
        original_scores = ifo.score_samples(tmp[[x, y]])
        method_scores['IsoFor'] = standardizer(original_scores.reshape(-1,1))

    if algo=='comb':
        preds  = np.where(preds2==-1, 1, 0) +\
                np.where(preds3==-1, 1, 0) +\
                np.where(preds4==-1, 1, 0)
        tmp['pred'] = np.where(preds>1, 1, 0)
    else:
        tmp['pred'] = np.where(preds==-1, 1, 0)

    for k, v in method_scores.items():
        return_scores[k] = v[:,0]
    
    return_scores = pd.DataFrame(return_scores)
    return_scores['agg'] = return_scores.mean(axis=1)
    anomal_idx.append(list(tmp[tmp.pred==1].index))
    return sum(anomal_idx, []), return_scores, tmp.index

def train_crossplot(**kwargs):
    pass

def get_proba_predictions(X, y, bad_logs, model, cp_name):
    """
    Add to the dataframe the score results for each of the bad anomalies.

    Args:
        X (pd.DataFrame): dataframe for prediction on
        y (pd.DataFrame): datafrrame with columns to populate results
        bad_logs (pd.DataFrame): bad log samples dataframe
        models (dict): dictionary with models per cp
        cp_name (str): crossplot name

    Returns:
        y: pd.DataFrame populated with results
    """
    cp_preds = model.predict_proba(X)
    if cp_name == 'vp_den':
        y.loc[bad_logs.index, 'score_vpden_ac']  = cp_preds[:, 0]
        y.loc[bad_logs.index, 'score_vpden_den'] = cp_preds[:, 1]
    elif cp_name == 'vp_vs':
        y.loc[bad_logs.index, 'score_vpvs_ac']  = cp_preds[:, 0]
        y.loc[bad_logs.index, 'score_vpvs_acs'] = cp_preds[:, 1]
    elif cp_name == 'ai_vpvs':
        y.loc[bad_logs.index, 'score_aivpvs_ac']  = cp_preds[:, 0]
        y.loc[bad_logs.index, 'score_aivpvs_acs'] = cp_preds[:, 1]
        y.loc[bad_logs.index, 'score_aivpvs_den'] = cp_preds[:, 2]
    return y

def get_class_predictions(X, y, bad_logs, model, cp_name):
    """
    Add to the dataframe the class results for each of the bad anomalies.

    Args:
        X (pd.DataFrame): dataframe for prediction on
        y (pd.DataFrame): datafrrame with columns to populate results
        bad_logs (pd.DataFrame): bad log samples dataframe
        models (dict): dictionary with models per cp
        cp_name (str): crossplot name

    Returns:
        y: pd.DataFrame populated with results
    """
    cp_preds = model.predict(X)
    y['tmp'] = 0
    y.loc[bad_logs.index, 'tmp'] = cp_preds
    #Map predictions depending on which crossplot they come from
    if cp_name == 'vp_den':
        y.loc[y.tmp.isin([1,3]), 'flag_vpden_ac']  = 1
        y.loc[y.tmp.isin([2,3]), 'flag_vpden_den'] = 1        
    elif cp_name == 'vp_vs':
        y.loc[y.tmp.isin([1,3]), 'flag_vpvs_ac']  = 1
        y.loc[y.tmp.isin([2,3]), 'flag_vpvs_acs'] = 1 
    elif cp_name == 'ai_vpvs':
        y.loc[y.tmp.isin([4,5,6,7]), 'flag_aivpvs_den'] = 1
        y.loc[y.tmp.isin([1,3,5,7]), 'flag_aivpvs_ac']  = 1
        y.loc[y.tmp.isin([2,4,6,7]), 'flag_aivpvs_acs'] = 1
    return y

def flag_well_crossplots(X, x, y, algo, cp_name, **kwargs):
    """
    Returns anomalous indices of crossplots between x and y

    Args:
        X (pd.DataFrame): data of one well

    Returns:
        tuple: dataframe with scores and lists of anomalous indices
    """
    res = dict(zip(
        ['anomalies', 'scores', 'idx'],
        find_crossplot_scores(X, x, y, algo=algo, **kwargs)
    ))
    for comb_method, scs in res['scores'].items():
        method_col = '{}_{}'.format(comb_method, cp_name)
        X.loc[res['idx'], method_col] = scs.values
        X[method_col].fillna(0, inplace=True)
    return X, res['anomalies']

def flag_crossplots(X, x, y, algo, cp_name, **kwargs):
    """
    Returns anomalous indices of crossplots between x and y

    Args:
        X (pd.DataFrame): data of one well

    Returns:
        tuple: dataframe with scores and lists of anomalous indices
    """
    anomalies = []
    for lsu in X['GROUP'].unique():
        lsu_data = X[X['GROUP']==lsu]
        res = dict(zip(
            ['anomalies', 'scores', 'idx'],
            find_crossplot_scores(lsu_data, x, y, algo=algo, **kwargs)
        ))
        for comb_method, scs in res['scores'].items():
            method_col = '{}_{}'.format(comb_method, cp_name)
            X.loc[res['idx'], method_col] = scs.values
            X[method_col].fillna(0, inplace=True)
        anomalies.extend(res['anomalies'])
    return X, anomalies

def classify_badlog(y, cp_name, ds, key_wells, model):
    bad_logs = ds.df_original[ds.df_original['flag_crossplot_gen']!=0]
    # if no badlogs, dont do anything
    if len(bad_logs) == 0:
        return y
    df_preprocessed, _, feats = ds.preprocess(
        ds.df_original.loc[bad_logs.index],
        _normalize_curves={'key_wells':key_wells}
    )
    #scores
    y_scores = get_proba_predictions(df_preprocessed[feats], y, bad_logs, model, cp_name)
    y[y_scores.columns] = y_scores
    #flags
    y_flags = get_class_predictions(df_preprocessed[feats], y, bad_logs, model, cp_name)
    y[y_flags.columns] = y_flags
    return y

def flag_crossplot(X_pred, cp_names, datasets, models, key_wells, **kwargs):
    print('Method: crossplot - general...')
    outlier_methods = ['EliEnv', 'SVM', 'IsoFor'] #FIXME! The shouldn't preferably be hard-coded.
    for cp_name in cp_names:        
        for method in outlier_methods:
            X_pred['{}_{}'.format(method, cp_name)] = 0.

    #get anomalies per crossplot and their scores

    vp_den_anomalies, ai_vpvs_anomalies, vp_vs_anomalies = [], [], []
    
    X_pred, vp_den_anomalies  = flag_crossplots(X_pred, 'VP', 'DEN', 'comb', 'vp_den', **kwargs)

    X_pred, ai_vpvs_anomalies = flag_crossplots(X_pred, 'AI', 'VPVS', 'comb', 'ai_vpvs', **kwargs)

    X_pred, vp_vs_anomalies   = flag_crossplots(X_pred, 'VP', 'VS', 'comb', 'vp_vs', **kwargs)

    y = X_pred.copy()        
    y.loc[:, [
        'flag_vpden_ac', 'flag_vpvs_ac', 'flag_aivpvs_ac',
        'flag_vpvs_acs', 'flag_aivpvs_acs',
        'flag_vpden_den', 'flag_aivpvs_den'
    ]] = 0, 0, 0, 0, 0, 0, 0

    pred_gen_cols = ['flag_crossplot_gen', 'flag_vpden_gen', 'flag_aivpvs_gen', 'flag_vpvs_gen']
    y.loc[:, pred_gen_cols] = 0, 0, 0, 0        
    y.loc[vp_den_anomalies, ['flag_crossplot_gen', 'flag_vpden_gen']]   = 1
    y.loc[ai_vpvs_anomalies, ['flag_crossplot_gen', 'flag_aivpvs_gen']] = 1
    y.loc[vp_vs_anomalies, ['flag_crossplot_gen', 'flag_vpvs_gen']]     = 1

    #Add the general crossplot flags to X_pred(to use as features in the model)
    X_pred[pred_gen_cols] = y[pred_gen_cols]

    #Run the anomalous samples through the relevant trained model and get predictions
    #Preprocessing using key_wells will need the well_name column, but the value is unimportant
    X_pred['well_name'] = 'dummy'
    for cp_name in cp_names:
        print('Method: crossplot - {}...'.format(cp_name))
        ds = datasets[cp_name]
        ds.load_from_df(X_pred)
        y = classify_badlog(y, cp_name, ds, key_wells[cp_name], models[cp_name])
    y = y.fillna(0)

    return y



from akerbp.models.rule_based_models import BS_helpers

def flag_bad_den_shallow(df_well):
    """
    Looks for parial DEN measurement for the same BS 
    in the shallow end of the well and mark those DEN measurements as bad
    Args:
        df_well (pd.DataFrame): well data

    Returns:
        list: indices with bad DEN
    """    
    df_well = BS_helpers.find_BS_jumps(df_well)
    #get the index of the first DEN value
    first_DEN_idx     = df_well[df_well['DEN'].notna()].index.min()
    first_BS_region   = df_well.loc[first_DEN_idx, 'BS_region']
    first_region_idx  = df_well[df_well['BS_region']==first_BS_region].index.min()
    bad_region_length = len(df_well[df_well['BS_region']==first_BS_region]['DEN'].dropna())
    if (first_region_idx < first_DEN_idx) and (bad_region_length<1000):
        dencorr_casing = df_well[df_well['BS_region']==first_BS_region].index
    else:
        dencorr_casing = []
    return dencorr_casing

def flag_bad_den_at_jump(df_well):
    """
    Looks around the regions where BS value changes, if there are missing DEN measurements, 
    marks X measurements before and after that as potentioally bad

    Args:
        df_well (pd.DataFrame): well data

    Returns:
        list: indices with bad DEN
    """    
    df_well = BS_helpers.find_BS_jumps(df_well)
    bad_den_idx = []
    for region in df_well['BS_region'].unique():
        min_region_idx = df_well[df_well['BS_region']==region].index.min()
        min_den_idx    = df_well[df_well['BS_region']==region]['DEN'].dropna().index.min()
        if min_den_idx > min_region_idx:
            bad_den_idx.extend(df_well[
                df_well['BS_region']==region]['DEN'].dropna().head(20).index.tolist()
            )
        max_region_idx = df_well[df_well['BS_region']==region].index.max()
        max_den_idx = df_well[df_well['BS_region']==region]['DEN'].dropna().index.max()
        if max_den_idx < max_region_idx:
            bad_den_idx.extend(df_well[
                df_well['BS_region']==region]['DEN'].dropna().tail(20).index.tolist()
            )
    return bad_den_idx

def flag_casing(df_well, y_pred=None):
    """
    Returns anomalous DEN due to invalid measurements inside the casing
    combining the findings of flag_bad_den_shallow() and flag_bad_den_at_jump() function

    Args:
        df_well (pd.DataFrame): well data
        y_pred (pd.DataFrame): dataframe to append the flag columns ('flag_casing_gen' and 'flag_casing_den') to. 
        If set to None, the columns are added to a copy of df_well - default: None

    Returns:
        pd.DataFrame: y_pred
    """
    print('Method: casing...')
    bad_den_casing = []
    bad_den_casing.extend(flag_bad_den_shallow(df_well))
    bad_den_casing.extend(flag_bad_den_at_jump(df_well))
    if y_pred is None:
        y_pred = df_well.copy()
    y_pred.loc[:, ['flag_casing_gen', 'flag_casing_den']] = 0, 0
    y_pred.loc[bad_den_casing, ['flag_casing_gen', 'flag_casing_den']] = 1
    return y_pred


import xgboost
import joblib
from sklearn.preprocessing import LabelEncoder
from sklearn.utils import class_weight
import numpy as np

from akerbp.models import Model


class ClassificationModel(Model):
        
    def predict_proba(self, X):
        """
        Parameters
        ----------
            X: dict or pandas DataFrame with expected features

        Returns
        -------
            y: numpy array with prediction probabilities of each class
        """
        if self.model is None:
            m = ('Model must be trained or loaded from file '
                 'before predict_proba can be run')
            raise ValueError(m)
        self.X_pred = X
        self._validate_x_pred()

        if hasattr(self, '_validate_features'):
            self._validate_features()

        return self.model.predict_proba(self.X_pred)

    def _prepare_sample_weights(self, targets):
        if self.use_class_weights:
            weights = class_weight.compute_class_weight(
                'balanced', classes=np.unique(targets), y=targets.values
            )
            sw = targets.copy()
            classes = sw.unique()

            zip_iterator = zip(classes, weights)
            self.class_weights = dict(zip_iterator)

            sw = sw.replace(self.class_weights)
            self.sample_weights = sw.values
        else:
            self.sample_weights = np.ones_like(targets)


class XGBoostClassificationModel(ClassificationModel):
    def __init__(self, settings, model_path, load=True):
        self.settings = settings
        for key, val in settings.items():
            setattr(self, key, val)
        self.model_path = model_path
        self._handle_model_path()
        if load:
            self.model = joblib.load(self.model_file_path)
            print('Model successfully loaded from ', self.model_file_path)
            self.le = joblib.load(self.le_file_path)
            print('LabelEncoder successfully loaded from ', self.le_file_path)

    def _validate_features(self):
        """
        Check that the features in the model match the features provided in 
        the data (including their order)
        """
        self.model_features = self.model.get_booster().feature_names
        if sorted(self.model_features) != sorted(self.X_pred.columns.values):
            m = ('Error in provided features. '
                f'Expected: {list(self.model_features)}'
                f'Provided: {list(self.X_pred.columns)}')
            raise ValueError(m)
        self.X_pred = self.X_pred[self.model_features]

    def predict(self, X):
        return super().predict(X, validate_features=True).astype('int')

    def train(self, df, targets, **kwargs):
        ''' 
        Trains on the data given in the provided dataframe and targets. 
        The curves of the model must be present in the dataframe. 

        Parameters
        ----------
            df:         Pandas dataframe of features
            targets:    List of integers of the same length as the number of 
                        rows in df
        '''
        self._prepare_sample_weights(targets)
        self.le = LabelEncoder()
        targets = self.le.fit_transform(targets)
        self.model = xgboost.XGBRFClassifier(**self.model_parameters)
        self.model.fit(df, targets, sample_weight=self.sample_weights, **kwargs)








import numpy as np

### BS-related methods
def _number_BS_regions(df_):
    """
    Gives each BS region a number

    Args:
        df (pd.DataFrame): should include 'BS_jump' column (others will be ignored)

    Returns:
        df (pd.DataFrame): same as the input dataframe with the added "BS_region" column with dtype='int'
    """
    df = df_.copy()
    start_idx = df.index.min()
    df['BS_region'] = np.nan
    for rid, idx in enumerate(df[df['BS_jump']==1].index):
        end_idx = idx
        df.loc[range(start_idx, end_idx), 'BS_region'] = rid
        start_idx = end_idx
    df['BS_region'].fillna(0, inplace=True)
    return df

def find_BS_jumps(df_):
    """
    Finds points where BS changes

    Args:
        df (pd.DataFrame): should include 'BS' column (others will be ignored)

    Returns:
        df (pd.DataFrame): same as the input dataframe, with the added 'BS_jump' with dtype='bool'
    """
    df = df_.copy()
    #First find points where BS changes
    df['BS_jump'] = 0
    df.loc[df['BS'].diff(1).ne(0), 'BS_jump'] = 1
    #mark each BS region with an ID
    df = _number_BS_regions(df)
    return df

def flag_BS(df_well, y_pred=None, **kwargs):
    """
    Returns anomalous BS values

    Args:
        df_well (pd.DataFrame): [description]

    Returns:
        [type]: [description]
    """
    print('Method: bitsize...')
    bs_step_size = kwargs.get('bs_step_size', 10)
    df_well      = find_BS_jumps(df_well)
    bs_anomalies = []
    for _, v in df_well[df_well['BS_jump'] == 0].groupby((df_well['BS_jump'] != 0).cumsum()):
        if v.shape[0] < bs_step_size:
            bs_anomalies.extend(v.index.tolist())
    if y_pred is None:
        y_pred = df_well.copy()
    y_pred.loc[:, ['flag_bitsize_gen']] = 0
    y_pred.loc[bs_anomalies, ['flag_bitsize_gen']] = 1
    return y_pred



import numpy as np
from akerbp.models.rule_based_models import helpers
import akerbp.models.rule_based_models.crossplot_helpers as crossplot_helpers

def flag_outliers(df_well, y_pred=None, **kwargs):
    """
    Returns anomalous indices for AC, ACS and DEN based on outliers of crossplots between AC, ACS and DEN
    and other curves

    Args:
        df_well (pd.DataFrame): data from one well

    Returns:
        tuple: lists of anomalous indices of AC, ACS and DEN
    """
    print('Method: outliers...')
    if 'CALI_BS' not in df_well.columns:
        df_well['CALI_BS'] = np.abs(df_well['CALI'] - df_well['BS'])
    if 'RDEP_log' not in df_well.columns and 'RDEP' in df_well.columns:
        df_well['RDEP_log'] = np.log10(df_well.RDEP)
    default_curves = ['AC', 'ACS', 'CALI', 'GR', 'NEU', 'PEF', 'RDEP_log', 'CALI_BS']
    curves = kwargs.get('curves', default_curves)

    if 'GROUP' in curves:
        curves.remove('GROUP')

    algo_params = {
        'DBSCAN_eps': 0.4,
        'EliEnv_contamination': 0.01,
        'EliEnv_random_state': 0,
        'SVM_gamma': 'scale',
        'SVM_nu': 0.01,
        'IsoFor_n_estimators': 50,
        'IsoFor_contamination': 0.005,
        'IsoFor_random_state': 0
    }
    if y_pred is None:
        y_pred = df_well.copy()

    ac_outliers_anomalies = []
    acs_outliers_anomalies = []
    den_outliers_anomalies = []

    logname_curves = {
        'AC': ([c for c in curves if c not in ['VP', 'AC']], ac_outliers_anomalies),
        'ACS': ([c for c in curves if c not in ['VS', 'ACS']], acs_outliers_anomalies),
        'DEN': ([c for c in curves if c!='DEN'], den_outliers_anomalies)
        }

    method = 'outliers'
    for logname, (curves, anomalies) in logname_curves.items():
        for y in curves:
            flags, scores, idx  = crossplot_helpers.find_crossplot_scores(df_well, x=logname, y=y, **algo_params)
            anomalies.extend(flags)
            for a_method in scores.keys():
                y_pred.loc[idx, '{}_{}_{}'.format(a_method, method, logname.lower())] = scores[a_method]
                y_pred['{}_{}_{}'.format(a_method, method,  logname.lower())].fillna(0, inplace=True)

    y_pred.loc[:, ['flag_outliers_gen', 'flag_outliers_ac', 'flag_outliers_acs', 'flag_outliers_den']] =\
        0, 0, 0, 0
    y_pred.loc[ac_outliers_anomalies, 'flag_outliers_ac']   = 1
    y_pred.loc[acs_outliers_anomalies, 'flag_outliers_acs'] = 1
    y_pred.loc[den_outliers_anomalies, 'flag_outliers_den'] = 1

    y_pred['flag_outliers_ac']  = helpers.fill_holes(y_pred, 'flag_outliers_ac')
    y_pred['flag_outliers_acs'] = helpers.fill_holes(y_pred, 'flag_outliers_acs')
    y_pred['flag_outliers_den'] = helpers.fill_holes(y_pred, 'flag_outliers_den')

    y_pred.loc[((y_pred.flag_outliers_ac==1) | (y_pred.flag_outliers_acs==1) |\
        (y_pred.flag_outliers_den==1)), 'flag_outliers_gen']  = 1

    return y_pred



import numpy as np
from akerbp.models.rule_based_models import helpers
from akerbp.models.rule_based_models import BS_helpers
from pyod.utils.utility import standardizer #returns z-score

def get_outliers(df_well, y_pred=None):
    
    """
    Returns CALI-BS outliers
    
    Args:
        df_well (pd.DataFrame): well data
        y_pred (pd.DataFrame): dataframe with flags
    
    Returns:
        pd.DataFrame: y_pred with added flag columns
    """ 
    if 'CALI' in df_well.columns and 'BS' in df_well.columns:
        df_well.loc[:, 'CALI_BS'] = np.abs(df_well['CALI'] - df_well['BS'])

    x = 'CALI_BS'
    y = 'CALI_BS'
    algo_params = {
        'DBSCAN_eps': 0.02,
        'EliEnv_contamination': 0.01,
        'EliEnv_random_state': 0,
        'SVM_gamma': 'scale',
        'SVM_nu': 0.01,
        'IsoFor_n_estimators': 20,
        'IsoFor_contamination': 0.05,
        'IsoFor_random_state': 0
    }

    df_well = BS_helpers.find_BS_jumps(df_well)

    if y_pred is None:
        y_pred = df_well.copy()

    # return if too many BS regions (a couple of wells with non-standard values)
    if df_well['BS_region'].nunique() >= 10:
        return y_pred
    
    y_pred.loc[:, ['flag_washout_gen', 'flag_washout_den']] = 0, 0
    den_washout_anomalies = []

    for bsr in df_well['BS_region'].unique():
        df_gr = df_well[df_well['BS_region']==bsr].copy()
        if (df_gr[x].dropna().shape[0] != 0) and (df_gr[y].dropna().shape[0] != 0):
            y_scores = helpers.get_crossplot_scores(
                df_gr, 'CALI_BS', ['CALI_BS'], y_pred, 'washout', **algo_params   
            )
            y_pred[y_scores.columns] = y_scores
                    
            den_washout_anomalies.extend(
                helpers.get_crossplot_anomalies(
                    df_gr, 'CALI_BS', ['CALI_BS'], **algo_params   
                )
            )
    y_pred.loc[den_washout_anomalies, ['flag_washout_gen', 'flag_washout_den']] = 1

    return y_pred
        
def flag_washout(df_well, y_pred=None, **kwargs):
    """
    Returns anomalous CALI-BS

    Args:
        df_well (pd.DataFrame): well data 
        y_pred (pd.DataFrame): dataframe to add detection columns to

    Returns:
        pd.DataFrame: y_pred with the added columns
    """
    print('Method: washout...')

    y_pred = get_outliers(df_well, y_pred)
    if not 'score_washout_gen' in y_pred.columns:
        y_pred.loc[:, 'score_washout_gen'] = 0
    rolling_window = kwargs.get('washout_rolling_window', 101)
    for bsr in y_pred['BS_region'].unique():
        df_gr = y_pred[y_pred['BS_region']==bsr].copy()
        rows_with_values = df_gr[df_gr['CALI'].isna()==False].index
        if len(rows_with_values):
            df_gr.loc[rows_with_values, 'CALI_std'] =\
                df_gr.loc[rows_with_values, 'CALI'].rolling(rolling_window).std().fillna(0)
            y_pred.loc[rows_with_values, 'score_washout_gen'] =\
                np.abs(standardizer(df_gr.loc[rows_with_values, 'CALI_std'].values.reshape(-1,1)))

    y_pred['score_washout_gen'].fillna(0, inplace=True)
    return y_pred


import numpy as np
from akerbp.models.rule_based_models import helpers
import akerbp.models.rule_based_models.crossplot_helpers as crossplot_helpers

### Log trends
def flag_logtrend(df_well, y_pred=None, **kwargs):
    """
    Returns anomalous indices for AC, ACS and DEN based on log trends

    Args:
        df_well (pd.DataFrame): data from one well
        y_pred (pd.DataFrame): output data to attach results to
        score (bool): if returned values should be score or binary

    Returns:
        tuple: lists of anomalous indices of AC, ACS and DEN
    """
    print('Method: logtrend...')
    if 'CALI_BS' not in df_well.columns:
        df_well['CALI_BS'] = np.abs(df_well['CALI'] - df_well['BS'])
    default_curves = ['CALI', 'GR', 'NEU', 'CALI_BS']
    curves_to_diff = kwargs.get('curves_to_diff', default_curves)
    if 'GROUP' in curves_to_diff:
        curves_to_diff.remove('GROUP')
    
    algo_params = {
        'DBSCAN_eps': 0.1,
        'EliEnv_contamination': 0.01,
        'EliEnv_random_state': 0,
        'SVM_gamma': 'scale',
        'SVM_nu': 0.005,
        'IsoFor_n_estimators': 50,
        'IsoFor_contamination': 0.005,
        'IsoFor_random_state': 0
    }
    
    #replace curves values by difference of consecutive samples
    tmp_curves_to_diff = curves_to_diff +\
        [c for c in ['AC', 'ACS', 'DEN'] if c not in curves_to_diff]
    df_well[tmp_curves_to_diff] = df_well[tmp_curves_to_diff].diff()
    
    if y_pred is None:
        y_pred = df_well.copy()

    ac_trends_anomalies = []
    acs_trends_anomalies = []
    den_trends_anomalies = []

    logname_curves = {
        'AC': ([c for c in curves_to_diff if c not in ['VP', 'AC']], ac_trends_anomalies),
        'ACS': ([c for c in curves_to_diff if c not in ['VS', 'ACS']], acs_trends_anomalies),
        'DEN': ([c for c in curves_to_diff if c!='DEN'], den_trends_anomalies)
        }

    method = 'logtrend'
    for logname, (curves, anomalies) in logname_curves.items():
        for y in curves:
            flags, scores, idx  = crossplot_helpers.find_crossplot_scores(df_well, x=logname, y=y, **algo_params)
            anomalies.extend(flags)
            for a_method in scores.keys():
                y_pred.loc[idx, '{}_{}_{}'.format(a_method, method, logname.lower())] = scores[a_method]
                y_pred['{}_{}_{}'.format(a_method, method,  logname.lower())].fillna(0, inplace=True)

    y_pred.loc[:, ['flag_logtrend_gen', 'flag_logtrend_ac', 'flag_logtrend_acs', 'flag_logtrend_den']] =\
        0, 0, 0, 0
    y_pred.loc[ac_trends_anomalies, 'flag_logtrend_ac']   = 1
    y_pred.loc[acs_trends_anomalies, 'flag_logtrend_acs'] = 1
    y_pred.loc[den_trends_anomalies, 'flag_logtrend_den'] = 1

    y_pred['flag_logtrend_ac']  = helpers.fill_holes(y_pred, 'flag_logtrend_ac')
    y_pred['flag_logtrend_acs'] = helpers.fill_holes(y_pred, 'flag_logtrend_acs')
    y_pred['flag_logtrend_den'] = helpers.fill_holes(y_pred, 'flag_logtrend_den')

    y_pred.loc[((y_pred.flag_logtrend_ac==1) | (y_pred.flag_logtrend_acs==1) |\
        (y_pred.flag_logtrend_den==1)), 'flag_logtrend_gen']  = 1

    return y_pred



import numpy as np
import sklearn
import matplotlib.pyplot as plt
import seaborn as sns
import os
import akerbp.models.rule_based_models.crossplot_helpers as crossplot_helpers

def _make_col_dtype_lists(df):
    """
    Returns lists of numerical and categorical columns

    Args:
        df (pd.DataFrame): dataframe with columns to classify

    Returns:
        tuple: lists of numerical and categorical columns
    """
    num_cols = set(df._get_numeric_data().columns)
    cat_cols = list(set(df.columns) - set(num_cols))
    return list(num_cols), cat_cols

def _apply_metadata(df, **kwargs):
    """
    Applies specified metadata to data

    Args:
        df (pd.DataFrame): dataframe to apply metadata to

    Returns:
        tuple: pd.Dataframe after applying metadata, list of numerical columns and list of categorical columns
    """
    num_cols, cat_cols = _make_col_dtype_lists(df)
    num_filler         = kwargs.get('num_filler', None)
    cat_filler         = kwargs.get('cat_filler', None)
    if num_filler is not None:
        df.loc[:, num_cols] = df[num_cols].replace(to_replace=num_filler, value=np.nan)
    if cat_filler is not None:
        df.loc[:, cat_cols] = df[cat_cols].replace(to_replace=cat_filler, value=np.nan)
    return df, num_cols, cat_cols 

def _validate_features(X_pred_features, expected_curves):
    """
    Checks that provided and expected features are the same
    """
    if not X_pred_features.issubset(expected_curves):
        missing_features = list(expected_curves - X_pred_features)
        if len(missing_features) > 0:
            raise ValueError("Following features are expected but missing: {}".format(missing_features))
    if not expected_curves.issubset(X_pred_features):
        missing_features = list(X_pred_features - expected_curves)
        if len(missing_features) > 0:
            raise ValueError("Following features are expected but missing: {}".format(missing_features))

def _create_features(X_pred):
    """
    Creates features necessary for algorithms
    """
    if 'VP' not in X_pred.columns:
        X_pred.loc[:,'VP'] = 304.8 / X_pred['AC']
    if 'VS' not in X_pred.columns:
        X_pred.loc[:,'VS'] = 304.8 / X_pred['ACS']
    if 'AI' not in X_pred.columns:
        X_pred.loc[:,'AI'] = X_pred['DEN'] * ((304.8 / X_pred['AC'])**2)
    if 'VPVS' not in X_pred.columns:
        X_pred.loc[:,'VPVS'] = X_pred['VP'] / X_pred['VS']
    return X_pred

def fill_holes(df_, flag_col, limit=3):
    """
    Fill holes/include adjacent points to anoamlies as anomalies

    Args:
        df_ (pd.DataFrame): dataframe with the column to fill holes
        flag_col (string): column of df to fill holes
        limit (int, optional): how many samples to include as anomalye adjacent to anomalies. Defaults to 3.

    Returns:
        list: list of new values
    """
    tmp = df_.copy()
    tmp['filled'] = tmp[flag_col].replace(
        0, np.nan
        ).fillna(
            method='ffill',
            limit=limit
        ).fillna(
            method='bfill',
            limit=limit
        ).replace(np.nan, 0)
    return tmp.filled.values

#FIXME! move to crossplot_helpers
def get_crossplot_scores(df_well, logname, curves, y_pred, method,  **algo_params):
    """
    Returns scores for each sample

    Args:
        df_well (pd.DataFrame): dataframe with data from one well
        logname (str): log to which all curves will the crossploted against
        curves (list): list of curves to analyse
        y_pred (pd.DataFrame): dataframe with results per sample

    Returns:
        pd.DataFrame: y_pred with extra columns of scores
    """
    for y in curves:
        _, scores, idx  = crossplot_helpers.find_crossplot_scores(df_well, x=logname, y=y, **algo_params)
        for a_method in scores.keys():
            y_pred.loc[idx, '{}_{}_{}'.format(a_method, method, logname.lower())] = scores[a_method]
            y_pred['{}_{}_{}'.format(a_method, method,  logname.lower())].fillna(0, inplace=True)
    return y_pred

#FIXME! move to crossplot_helpers
def get_crossplot_anomalies(df_well, logname, curves, **algo_params):
    """
    Returns bool for each sample indicating anomaly or not

    Args:
        df_well (pd.DataFrame): dataframe with data from one well
        logname (str): log to which all curves will the crossploted against
        curves (list): list of curves to analyse

    Returns:
        pd.DataFrame: y_pred with extra columns of scores
    """
    anomalies = []
    for y in curves:
        tmp, _, _  = crossplot_helpers.find_crossplot_scores(df_well, x=logname, y=y, **algo_params)
        anomalies.extend(tmp)
    return anomalies

def expand_flags(df_well, flag_col, expansion_size=3):
    """
    Expands the flagged samples by expansion_size in each direction

    Args:
        df_well (pd.DataFrame): The input dataframe
        flag_col (str): Name of the 

    Returns:
        df_well (pd.DataFrame): input dataframe with updated flag_col
    """
    #get all available indices in the well dataframe
    well_idx = df_well.index.tolist()

    #then the flagged rows
    flag_idx = df_well[df_well[flag_col]==True].index.tolist()

    #Then find the flagged indices in the well
    flag_idx_idx = [well_idx.index(i) for i in flag_idx]
    
    for i in range(1, expansion_size):
        #check that the before and after indices are available in the well
        before_flag_idx_i = [well_idx[x-i] for x in flag_idx_idx if ((x-i)>=0) and (well_idx[x-i] in well_idx)]
        after_flag_idx_i  = [well_idx[x+i] for x in flag_idx_idx if ((x+i)<len(well_idx)) and (well_idx[x+i] in well_idx)]
        
        #update the flag_col
        df_well.loc[before_flag_idx_i, flag_col] = 1
        df_well.loc[after_flag_idx_i, flag_col]  = 1
        
    return df_well

def print_metrics(true, pred, ax=None, print_values=False, title=None, fig_name=None, plot_dir=None):
    """
    Plot confusion matrix with metrics of bad logs detection

    Args:
        true (pd.Series): true values
        pred (pd.Series): predicted values
        ax (list, optional): axes indices. Defaults to None.
        print_values (bool, optional): specify if printing is required with confusion matrix. Defaults to False.
        title (str, optional): specifies title for confusion matrix. Defaults to None.
    """
    recall = sklearn.metrics.recall_score(true, pred)
    prec   = sklearn.metrics.precision_score(true, pred)
    f1sc   = sklearn.metrics.f1_score(true, pred)

    conf_matrix = sklearn.metrics.confusion_matrix(true, pred)

    LABELS = [False, True]
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(5,5))
    sns.heatmap(conf_matrix, xticklabels=LABELS, yticklabels=LABELS, annot=True,  annot_kws={"fontsize":14}, fmt="d", ax=ax);
    
    if title is None:
        title = 'Confusion matrix'
    if print_values:
        title = f'{title} \n Recall = {recall:.3f} \n Precision = {prec:.3f} \n F1-score = {f1sc:.2f}'
    ax.set_title(title, {'size':'15'})
    ax.set_ylabel('Label', {'size':'18'})
    ax.set_xlabel('Predicted class', {'size':'18'})
    plt.tight_layout()
    if plot_dir is not None:
        plt.savefig(os.path.join(plot_dir, '{}.jpg'.format(fig_name)), dpi=150)


import numpy as np
import pandas as pd
import umap as UMAP
from sklearn.preprocessing import MinMaxScaler

### UMAP
def umap3d(df_well, curves_umap):
    """
    Returns anomalous indices found by umap clustering.
    Not in use at the moment.

    Args:
        df_well (pd.DataFrame): data from one well
        curves_umap (list): list of curves to cluster

    Returns:
        list: list of anomalous indices
    """
    umap = UMAP.UMAP(n_components=3)
    df_well_umap = umap.fit_transform(MinMaxScaler().fit_transform(df_well[curves_umap]))
    df_well_umap = pd.DataFrame(df_well_umap, index=df_well.index, columns=['first', 'second', 'third'])
    tmp_X = df_well_umap[['first', 'second', 'third']]
    preds3 = OneClassSVM(gamma='scale', nu=0.001).fit_predict(tmp_X)
    preds4 = IsolationForest(n_estimators=50, contamination=0.001).fit_predict(tmp_X)
    preds  = np.where(preds3==-1, 1, 0) +\
                np.where(preds4==-1, 1, 0)
    tmp_X['pred'] = np.where(preds>0, True, False)
    return list(tmp_X[tmp_X.pred==True].index)

def flag_umap(df_well, y_pred=None, **kwargs):
    """
    Flags umap anomalies for AC, ACS and DEN

    Args:
        df_well ([type]): [description]

    Returns:
        [type]: [description]
    """
    print('Method: UMAP...')

    if y_pred is None:
        y_pred = df_well.copy()

    y_pred.loc[:, ['flag_umap_gen', 'flag_umap_ac', 'flag_umap_acs', 'flag_umap_den']] =\
        0, 0, 0, 0

    curves_umap = ['AC', 'GR', 'NEU', 'RDEP', 'RMED', 'VP', 'AI', 'VPVS']
    try:
        ac_umap = umap3d(df_well, curves_umap)    
    except:
        ac_umap = []  
    curves_umap = ['ACS', 'GR', 'NEU', 'RDEP', 'RMED', 'VS', 'AI', 'VPVS']
    try:
        acs_umap = umap3d(df_well, curves_umap)
    except:
        acs_umap = []
    curves_umap = ['GR', 'NEU', 'RDEP', 'RMED', 'AI', 'VPVS', 'DEN', 'DENC']
    try:
        den_umap = umap3d(df_well, curves_umap)
    except:
        den_umap = []

    y_pred.loc[ac_umap, ['flag_umap_gen', 'flag_umap_ac']]   = 1
    y_pred.loc[acs_umap, ['flag_umap_gen', 'flag_umap_acs']] = 1
    y_pred.loc[den_umap, ['flag_umap_gen', 'flag_umap_den']] = 1

    return y_pred


"""
setup.py 

Information used to build the package
"""
from setuptools import find_namespace_packages, setup
import os
import subprocess


with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()


def get_version():
    ENV = os.environ['ENV'] # Must be set
    tag = subprocess.check_output(
        ['git', 'describe', '--tags', '--exact-match'],
        encoding='UTF-8'
    ).rstrip()
    if ENV in ['dev', 'test']:
        version = '0.' + tag
    elif ENV == 'prod':
        version = '1.' + tag
    else:
        raise ValueError(f"Unknown environment {ENV}")
    return version


setup(
    name="akerbp.models", 
    version=get_version(),
    author="Alfonso M. Canterla",
    author_email="alfonso.canterla@soprasteria.com",
    description="Machine Learning Models for Petrophysics",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/akerbp/models/",
    packages=find_namespace_packages(where='src'),
    package_dir={'': 'src'},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=[
        "xgboost==1.3.3",
        "joblib==1.0.1",
        "numpy>=1.19.5",
        "pandas>=1.3.2",
        "scikit-learn>=0.24.2",
    ]
)


import numpy as np
import pandas as pd
from pandas.core.frame import DataFrame, Series
from typing import Tuple, List
from akerbp.models.rule_based_models import helpers


# Flatline anomalies
def get_sig_outliers(df: DataFrame, th: float = 2.5) -> Tuple[List, List]:
    """
    Returns indices of anomalous values of resistivities and scores
    representing how far from the mean (in standard deviations) they are.

    Args:
        df (pd.Series): series with data to analyse
        th (float, optional): multiplied by the number of standard deviations.
        Defaults to 3.

    Returns:
        tuple: list of anomalous indices and scores
    """
    mu = df.mean()
    std = df.std()
    outliers = list(df[df > mu + th * std].index)
    scores = abs((df - mu) / std).values.tolist()
    return outliers, scores


def get_resistivity_outliers(df_well: Series, method: str, th: float) -> Tuple[List, List]:
    """
    Returns outliers indices and scores for given series.

    Args:
        df_well (Series): series with values to find outliers
        method (str): get outliers from the derivative or variance
        th (float): number of standard deviations to consider an outlier

    Returns:
        Tuple: list of outliers indices and list of scores for all samples
    """
    if method == "derivative":
        df = pd.Series(np.abs(np.gradient(df_well)), index=df_well.index)
    elif method == "variance":
        df = pd.Series(
            df_well.rolling(window=5, center=True).var(), index=df_well.index
        )
    df_nona = df[df != 0].dropna()
    if len(df_nona) == 0:
        return [], np.zeros(len(df_well))
    outliers, scores = get_sig_outliers(df_nona, th=th)
    df_scores = pd.Series(np.zeros(len(df_well)), index=df_well.index)
    df_scores.loc[df_nona.index] = scores
    return outliers, df_scores.values


def flag_resistivities(df_well, y_pred=None, **kwargs):
    """
    Flag all resistivities outliers in the given data based on mean+th*std.

    Args:
        df_well (DataFrame): dataframe with resistivities columns
        y_pred (DataFrame, optional): dataframe with outputs to add new outputs.
        Defaults to None.

    Returns:
        DataFrame: dataframe with output columns
    """
    print("Method: resistivities...")
    cols = kwargs.get("cols", ["RDEP", "RMED", "RSHA", "RMIC"])
    cols = [c for c in cols if c in df_well.columns]
    if y_pred is None:
        y_pred = df_well.copy()
    for method in ["derivative", "variance"]:
        flag_cols = [f"flag_{method}_{c.lower()}" for c in cols]
        y_pred.loc[:, flag_cols] = 0
        for c in cols:
            sig_idx, scores = get_resistivity_outliers(
                df_well[c], method, th=2.7
            )
            y_pred.loc[sig_idx, [f"flag_{method}_{c.lower()}"]] = 1
            # include two samples before and after as anomalies as well
            y_pred[f"flag_{method}_{c.lower()}"] = helpers.fill_holes(
                y_pred,
                f"flag_{method}_{c.lower()}",
                limit=2
            )
            y_pred[f"agg_{method}_{c.lower()}"] = scores
    return y_pred



import os
import numpy as np

from akerbp.models.classification_models import ClassificationModel
from akerbp.models.classification_models import XGBoostClassificationModel

class NestedShaleModel(ClassificationModel):
    ''' A wrapper object for two models

        1. A model that predicts shale
        2. A model that predicts all classes except shale
    '''
    
    def __init__(self, shale_settings, other_settings, model_path):
        self.model_path = model_path
        self.shale_settings = shale_settings
        self.other_settings = other_settings
        self.shale_model = XGBoostClassificationModel(shale_settings, os.path.join(self.model_path, 'shale_model'))
        self.other_model = XGBoostClassificationModel(other_settings, os.path.join(self.model_path, 'other_model'))        
        self.label_shale = 65000

    def predict(self, X):
        if self.shale_model is None or self.other_model is None:
            raise ValueError('Model must be trained or loaded from file before predict can be run')
        shale_prediction = self.shale_model.predict(X)
        other_prediction = self.other_model.predict(X[shale_prediction == -1])
        full_prediction = np.zeros(X.shape[0])
        full_prediction[shale_prediction == self.label_shale] = self.label_shale
        full_prediction[shale_prediction == -1] = other_prediction
        return full_prediction.astype(int)

    def train(self, df, targets):
        # Train the binary shale classification model
        print('Training shale model..')
        targets_shale = np.copy(targets)
        targets_shale[targets_shale != self.label_shale] = -1 
        self.shale_model.train(df, targets_shale)
        print('Training model for non-shale lithologies..')
        targets_nonshale = targets[targets != self.label_shale].copy()
        df_nonshale = df[targets != self.label_shale]
        self.other_model.train(df_nonshale, targets_nonshale)

    def save(self, model_path=None):
        if model_path is None:
            model_path = self.model_path

        self.shale_model.save(model_path=os.path.join(self.model_path, 'shale_model'))
        self.other_model.save(model_path=os.path.join(self.model_path, 'other_model'))



import numpy as np

### Flatline anomalies
def get_constant_derivatives(df, logname, th=0.0005):
    """
    Returns indices of anomalous samples based on derivatives lower than a threshold

    Args:
        df (pd.DataFrame): dataframe with data to analyse
        logname (str): curve to be flagged, if any
        th (float, optional): threshold to which lower values will be considered constant. Defaults to 0.0005.

    Returns:
        list: indices of constant derivatives
    """
    df['d1']              = np.gradient(df[logname], df['DEPTH'])
    df['d2']              = np.gradient(df['d1'], df['DEPTH'])
    df['derivative']      = np.abs(df['d1']-df['d2'])
    df['cnst_derivative'] = df['derivative']<th
    return df[df['cnst_derivative']==True].index, df['derivative']

def get_constant_windows(df, logname, window_size=10, min_periods=5, th=0.0005):
    """
    Returns indices of anomalous samples based on derivatives lower than a threshold within a window

    Args:
        df (pd.DataFrame): dataframe with data to analyse
        logname (str): curve to be flagged, if any
        window_size (int): size of window
        min_periods (int): min_periods of rolling window
        th (float, optional): threshold to which lower values will be considered constant. Defaults to 0.0005.

    Returns:
        list: indices of constant derivatives
    """
    df['minw']        = df[logname].rolling(window_size, min_periods=min_periods, center=True).min()
    df['maxw']        = df[logname].rolling(window_size, min_periods=min_periods, center=True).max()
    df['window_derivative'] = np.abs(df['maxw']-df['minw'])
    df['cnst_window'] = df['window_derivative']<th
    return df[df['cnst_window']==True].index, df['window_derivative']

def get_den_flatlines(df, th=0.002):
    """
    Returns flatlines specifically for density, where threshold was decided empirically

    Args:
        df (pd.DataFrame): dataframe with data to analyse
        th (float, optional): threshold to which lower values will be considered constant. Defaults to 0.002.

    Returns:
        list: indices of anomalous density samples
    """
    if len(df['DEN'])>1:
        df['DEN_flat'] = np.gradient(df['DEN'])
        idx = df[
            (np.abs(df['DEN_flat'].diff())<th) &\
            (np.abs(df['DEN_flat'].diff(-1))<th)
        ].index
    else:
        idx = []
    return idx

def get_flatlines(df, logname, cols, window=5, var_window=10, n_cols=2):
    """
    Returns indices of anomalous values indicating flatline badlog

    Args:
        df (pd.DataFrame): data to analyze
        logname (str): logname to find anomalies
        cols (list): list of curves to analyse logname against
        window (int, optional): window size for window constant derivatives. Defaults to 10.

    Returns:
        list: indices of flatlines anomlies
    """
    # get gradients and see if one gradient is not as small as at least three other curves
    cols_grad     = [c+'_gradient' for c in cols]
    gradient_flag = (df[f'{logname}_gradient']<df[f'{logname}_gradient'].rolling(var_window, center=True).std()) &\
                    (np.sum(df[cols_grad]>df[cols_grad].rolling(var_window, center=True).std(), axis=1)>n_cols)
    # get variance
    variance      = df[[logname]+cols].rolling(window).var()
    variance_flag = (variance[logname]<variance[logname].rolling(var_window, center=True).std()) &\
                    (np.sum(variance[cols]>variance[cols].rolling(var_window, center=True).std(), axis=1)>n_cols)
    # flatlines
    flatlines = gradient_flag | variance_flag
    return flatlines[flatlines==True].index

def find_start_end_indices(df_well):
    min_idx = df_well.index.max()
    max_idx = df_well.index.min()
    for col in ['AC', 'ACS', 'DEN']:
        min_idx = min(min_idx, df_well[df_well[col].notna()].index.min())
        max_idx = max(max_idx, df_well[df_well[col].notna()].index.max())
    return min_idx, max_idx

def flag_flatline(df_well, y_pred=None, **kwargs):
    """
    Returns flatlines indices for AC, ACS and DEN

    Args:
        df_well (pd.DataFrame): data from one well

    Returns:
        tuple: lists of gradient anomalies for AC, ACS and DEN
    """
    print('Method: flatline...')
    expected_curves = kwargs.get('expected_curves', set(['DEN', 'AC', 'ACS', 'GR', 'NEU', 'RDEP', 'RMED']))
    cols            = kwargs.get('cols', expected_curves)
    ncols           = kwargs.get('ncols', 2)

    for col in cols:
        df_well.loc[:, col+'_gradient'] = np.abs(np.gradient(df_well[col]))

    def run_get_flatlines(df_well, cols, ncols, **kwargs):
        window_size     = kwargs.get('window', 10)
        var_window_size = kwargs.get('var_window', 20)
        ac_grad_anomalies = get_flatlines(
            df=df_well, 
            logname='AC', 
            cols=list(set(cols) - set(['AC'])), 
            window=window_size,
            var_window=var_window_size,
            n_cols=ncols
        )
        acs_grad_anomalies = get_flatlines(
            df=df_well, 
            logname='ACS',
            cols=list(set(cols) - set(['ACS'])), 
            window=window_size,
            var_window=var_window_size,
            n_cols=ncols
        )
        den_grad_anomalies = get_flatlines(
            df=df_well, 
            logname='DEN', 
            cols=list(set(cols) - set(['DEN'])), 
            window=window_size,
            var_window=var_window_size,
            n_cols=ncols
        )
        return ac_grad_anomalies.tolist(), acs_grad_anomalies.tolist(), den_grad_anomalies.tolist()

    ac_grad_anomalies  = []
    acs_grad_anomalies = []
    den_grad_anomalies = []

    ac_flat_grad, ac_derivatives   = get_constant_derivatives(df_well, 'AC')
    acs_flat_grad, acs_derivatives = get_constant_derivatives(df_well, 'ACS')
    den_flat_grad, den_derivatives = get_constant_derivatives(df_well, 'DEN')

    ac_grad_anomalies.extend(ac_flat_grad)
    acs_grad_anomalies.extend(acs_flat_grad)
    den_grad_anomalies.extend(den_flat_grad)
    
    ac_flat_window, ac_window_grad   = get_constant_windows(df_well, 'AC')
    acs_flat_window, acs_window_grad = get_constant_windows(df_well, 'ACS')
    den_flat_window, den_window_grad = get_constant_windows(df_well, 'DEN')

    ac_grad_anomalies.extend(ac_flat_window)
    acs_grad_anomalies.extend(acs_flat_window)
    den_grad_anomalies.extend(den_flat_window)

    gen_grad_anomalies = []
    gen_grad_anomalies.extend(ac_grad_anomalies)
    gen_grad_anomalies.extend(acs_grad_anomalies)
    gen_grad_anomalies.extend(den_grad_anomalies)

    #run the function on points where there was constant window or gradient flags
    ac_grad_anomalies_main, acs_grad_anomalies_main, den_grad_anomalies_main = run_get_flatlines(
        df_well.loc[gen_grad_anomalies], cols, ncols, **kwargs)
    
    den_grad_anomalies_main.extend(get_den_flatlines(df_well.loc[gen_grad_anomalies]))
    
    #pick the start and end of well
    start_end_length = 50
    start_end_window_size = 2
    start_end_var_window_size = 2
    min_idx, max_idx = find_start_end_indices(df_well)
    start_condition = (df_well.index>=min_idx) & (df_well.index<=min_idx+start_end_length)
    end_condition = (df_well.index>=max_idx-start_end_length) & (df_well.index<=max_idx)

    #run the function for end points only with smaller window sizes
    df_well_ends = df_well[start_condition | end_condition]
    ac_grad_anomalies_ends, acs_grad_anomalies_ends, den_grad_anomalies_ends = run_get_flatlines(
        df_well_ends,
        cols,
        ncols,        
        **{'window': start_end_window_size, 'var_window': start_end_var_window_size}
        )

    ac_grad_anomalies_main.extend(ac_grad_anomalies_ends)
    acs_grad_anomalies_main.extend(acs_grad_anomalies_ends)
    den_grad_anomalies_main.extend(den_grad_anomalies_ends)

    if y_pred is None:
        y_pred = df_well.copy()

    y_pred.loc[:, ['flag_flatline_gen', 'flag_flatline_ac', 'flag_flatline_acs', 'flag_flatline_den']] =\
        0, 0, 0, 0

    #DEBUG   
    flag_columns = ['flag_den_only']
    lognames = ['ac', 'acs', 'den']
    for logname in lognames:
        flag_columns.extend([
            'flag_main_{}'.format(logname), 
            'flag_ends_{}'.format(logname), 
            'flag_derivs_{}'.format(logname), 
            'flag_windows_{}'.format(logname)
            ])
    y_pred.loc[:, flag_columns] = 0

    y_pred.loc[ac_grad_anomalies_main, ['flag_flatline_gen', 'flag_flatline_ac']]   = 1
    y_pred.loc[acs_grad_anomalies_main, ['flag_flatline_gen', 'flag_flatline_acs']] = 1
    y_pred.loc[den_grad_anomalies_main, ['flag_flatline_gen', 'flag_flatline_den']] = 1

    return y_pred


import os
import joblib
import xgboost
import pandas as pd

from akerbp.models import Model

class RegressionModel(Model):
    pass

class XGBoostRegressionModel():
    """    
    subclass of akerbp.model

    Attributes
    ----------
        settings: dict
            the required keys for the dictionary:
                arbitrary model name: dict containing following keys (dict):
                    'model_parameters': dict of parameters required by xgboost.XGBRFClassifier (dict),
                    'use_class_weights': whether to use class weights in the model (boolean)
        model_path: str
            - If model_path is a path/to/dir/ then the serialized model is assumed to be (or will be saved to) 
            path/to/dir/model.joblib, and the labedl encoder at /path/to/dir/le.joblib
            - If model_path is a full/path/to/my_model.joblib,
            then the label encoder is saved (or assumed to be) at /full/path/to/my_model_le.joblib

    Methods
    -------
        predict(X:dict/pandas.DataFrame)

        predict_proba(X:dict/pandas/DataFrame)

        train(df:pandas.DataFrame, targets:np.array)

        evaluate(X_test:pandas.DataFrame, y_test:pandas.DataFrame, metrics:list, label_map:dict, **kwargs)

        save(model_path:str)


    """
    def __init__(self, settings, model_path, load=True):

        self.settings = settings

        for key, val in settings.items():
            setattr(self, key, val)
                
        self.model_path = model_path
        self._handle_model_path()
        
        if load:
            try:
                self.model = joblib.load(self.model_file_path)
                print('Model successfully loaded from ', self.model_file_path)
            except:
                pass

    def _handle_model_path(self):
        if os.path.isfile(self.model_path):
            self.model_file_path = self.model_path
        else:
            if not os.path.isdir(self.model_path):
                os.makedirs(self.model_path)
            self.model_file_path = os.path.join(self.model_path, 'model.joblib')

    def _handle_model_path(self):
        if os.path.isfile(self.model_path):
            self.model_file_path = self.model_path
        else:
            if not os.path.isdir(self.model_path):
                os.makedirs(self.model_path)
            self.model_file_path = os.path.join(self.model_path, 'model.joblib')

    def _validate_features(self):
        # This function should be able to take in features in their
        # raw, unprocessed form as read from the file test.csv and
        # return predictions as an array integers of the same length
        if self.X_pred is None:
            return []

        #Check that the features in the model match the features provided in the data (including their order)
        self.model_features = self.model.get_booster().feature_names
        if sorted(self.model_features) != sorted(self.X_pred.columns.values):
            raise ValueError('Features in the provided dataset does not match the model expectation - Expected: {}, Provided: {}'.format(
        list(self.model_features), list(self.X_pred.columns)))
        self.X_pred = self.X_pred[self.model_features]

    def predict(self, X):
        """
        Parameters
        ----------
            X: dict or pandas DataFrame with expected features

        Returns
        -------
            y: numpy array with predicted classes
        """
        if isinstance(X, dict):
            self.X_pred = pd.DataFrame.from_dict(X)
        elif isinstance(X, pd.core.frame.DataFrame):
            self.X_pred = X
        else:
            raise ValueError('Please pass the data as a dict or a pandas DataFrame')
        self._validate_features()
        return self.model.predict(self.X_pred, validate_features=True)

    def predict_proba(self, X):
        """
        Not relevant for regression model
        """
        raise ValueError('predict_proba only relevant for classification models')

    def train(self, df, targets):
        ''' 
        Trains on the data given in the provided dataframe and targets. 
        The curves of the model must be present in the dataframe. 

        Parameters
        ----------
            df:         Pandas dataframe of features
            targets:    List of integers of the same length as the number of rows in df
        '''

        self.model = xgboost.XGBRegressor(**self.model_parameters)
        self.model.fit(df, targets)

    #FIXME!
    def explain(self):
        raise ValueError('Not implemented for this model')

    def evaluate(self, X_test, y_test, metrics=None, label_map=None, **kwargs):
        raise ValueError('Not implemented for this method')

    def save(self, model_path=None):
        if model_path is not None:
            self.model_path = model_path            

        self._handle_model_path()
        joblib.dump(self.model, self.model_file_path)
        print('Model successfully saved to {}'.format(self.model_file_path))
        return model_path

